  #include <stdlib.h> /* malloc/free */
  #include <assert.h> /* assertions */
  #include <string.h> /* strlen, strcpy */
  #include "wordtree.h" /* the prototypes */

  typedef struct treenode_struct {
    struct treenode_struct* pLeft;
    struct treenode_struct* pRight;
    char* word;
    void* userdata;
  } TREENODE;

  struct wordtree_struct {
    TREENODE* pRoot;
    wordtree_comparison comp;
  };

  static void splay(WORDTREE* tree,
		    const char* word);

  /* Helper function for wordtree_insert and         */
  /* wordtree_find. Top down Splay the tree,         */
  /* as described in Sleator's paper, searching for  */
  /* word. If the word is found, the tree root will  */
  /* point to that node.                             */


  static void subtreedestroy(TREENODE* node,
                             wordtree_userdatadel del);
  /* Helper function for destruction. It traverses   */
  /* tree and calls del (if not NULL) for the        */
  /* userdata of every node, before deallocating the */
  /* node itself.                                    */


  static int subtreetraverse(TREENODE* node,
                             wordtree_wordvisitor visitor,
			     void* always);
  /* Helper function for wordtree_traverse.          */
  /* Traverses the entire tree from node and         */
  /* downwards and calls visitor in sorted order     */


  WORDTREE* wordtree_create(wordtree_comparison comp)
  {
    WORDTREE* result=(WORDTREE*)malloc(sizeof(WORDTREE));
    if (result != NULL)
    {
      result->pRoot = NULL; /* initialise empty tree */
      result->comp = comp;  /* with comparison func. */
    }
    return result;
  }


  void** wordtree_insert(WORDTREE* tree, 
                         const char* word)
  {
    TREENODE* theNode;
    TREENODE* newNode;

    int compresult;
    /* Preconditions:                              */
    /*   tree != NULL                              */
    assert(tree != NULL);

    /*   word != NULL                              */
    assert(word != NULL);

    splay(tree, word);

    compresult = 0;

    if (tree->pRoot != NULL) {
      theNode = tree->pRoot;
      compresult = tree->comp(word, theNode->word);
      if (compresult == 0) /* found it */
	return &(theNode->userdata);
    }

    newNode = (TREENODE*)malloc(sizeof(TREENODE));
    if (newNode == NULL)
      return NULL;

    newNode->userdata = NULL;
    newNode->word = (char*)malloc(strlen(word)+1);
    strcpy(newNode->word, word);
    
    if (theNode == NULL) {
      newNode->pLeft = NULL;
      newNode->pRight = NULL;
    } else if (compresult < 0) {
      newNode->pLeft = theNode->pLeft;
      newNode->pRight = theNode;
      theNode->pLeft = NULL;
    } else {
      newNode->pRight = theNode->pRight;
      newNode->pLeft = theNode;
      theNode->pRight = NULL;
    }

    tree->pRoot = newNode;
    return &(newNode->userdata);
  }
 

  void** wordtree_find(WORDTREE* tree,
                      const char* word)
  {
    /* Preconditions:                              */
    /*   tree != NULL                              */
    assert(tree != NULL);

    /*   word != NULL                              */
    assert(word != NULL);

    splay(tree, word);

    if (tree->pRoot == NULL || tree->comp(word, tree->pRoot->word) != 0)
      return NULL;

    return &(tree->pRoot->userdata);
  }


  int wordtree_traverse(WORDTREE* tree,
                        wordtree_wordvisitor visitor,
			void* always)
  {    
    /* Preconditions:                               */
    /*   tree != NULL                               */
    assert(tree != NULL);

    /*   visitor != NULL                            */
    assert(visitor != NULL);

    return subtreetraverse(tree->pRoot, visitor, always);
  }


  void wordtree_destroy(WORDTREE* tree,
                        wordtree_userdatadel deletor)
  {
    /* Preconditions:                         */
    /*   tree != NULL                         */
    assert(tree != NULL);

    /* Destroy all nodes in the tree, and the */
    /* associated user data (if applicable)   */
    subtreedestroy(tree->pRoot, deletor);

    /* Then deallocate the wordtree struct    */
    free(tree);
  }

  /********************/
  /* Helper functions */
  /*******************/

  static void splay(WORDTREE* tree,
		    const char* word)
  { /* See description in beginning of file */

    TREENODE* node = tree->pRoot;
    TREENODE N;
    TREENODE* l;
    TREENODE* r;
    TREENODE* y;

    if (node == NULL)
      return;

    N.pLeft = NULL;
    N.pRight = NULL;
    l = &N;
    r = &N;

    for (;;) {
      int result = tree->comp(word, node->word);
      if ( result < 0)  {
	if (node->pLeft == NULL)
	  break;

	if (tree->comp(word, node->pLeft->word) < 0) {
	  y = node->pLeft;
	  node->pLeft = y->pRight;
	  y->pRight = node;
	  node = y;
	  if (node->pLeft == NULL)
	    break;
	}
	r->pLeft = node;
	r = node;
	node = node->pLeft;
      } else {
	if (result > 0) {
	  if (node->pRight == NULL)
	    break;
	  
	  if (tree->comp(word, node->pRight->word) > 0) {
	    y = node->pRight;
	    node->pRight = y->pLeft;
	    y->pLeft = node;
	    node = y;
	    if (node->pRight == NULL)
	      break;
	  }
	  l->pRight = node;
	  l = node;
	  node = l->pRight;
	} else
	  break;
      }
    }
    l->pRight = node->pLeft;
    r->pLeft = node->pRight;
    node->pLeft = N.pRight;
    node->pRight = N.pLeft;
    tree->pRoot = node;
  }

  static void subtreedestroy(TREENODE* node,
                             wordtree_userdatadel del)
  { /* See description in beginning of file, and */
    /* detailed description of tree traversal    */
    /* further down.                             */
    if (node != NULL)
    {
      /* First destroy everything to the left */
      subtreedestroy(node->pLeft, del);

      /* Then everything to the right */
      subtreedestroy(node->pRight, del);

      /* Then the user data, if del allows us */
      if (del != NULL)
	{
	  del(node->userdata);
	}

      /* And then deallocate this node */
      free(node);
    }
  }

  static int subtreetraverse(TREENODE* node,
                             wordtree_wordvisitor visitor,
			     void* commondata)
  { /* See description in beginning of file, and */
    /* detailed description of tree traversal    */
    /* further down.                             */
    if (node == NULL)
    {
      return 1;
    }
    return subtreetraverse(node->pLeft,
			   visitor,
			   commondata)
        && visitor(node->word,
		   node->userdata,
		   commondata)
        && subtreetraverse(node->pRight,
			   visitor,
			   commondata);

    /* Here we make use of C's lazy evaluation of */
    /* logical expressions.                       */
    /*                                            */
    /* If the first call to subtreetraverse       */
    /* returns 0, the whole expression will be    */
    /* zero, so nothing more needs to be done. In */
    /* this case visitor and the second           */
    /* subtreetraverse will not be called.        */
    /* The same happens if visitor returns 0,     */
    /* then the whole expression will be zero     */
    /* anyway, so the last subtreetraverse will   */
    /* not be called.                             */
    /*                                            */
    /* In other words, if any of the above        */
    /* functions returns 0, the rest of them will */
    /* not be called (thus aborting the           */
    /* traversal) and the function will return 0. */
    /* This zero return value will then be        */
    /* propagated upwards, immediately terminating*/
    /* all traversals.                            */
  }
