struct wordtree_struct;
typedef struct wordtree_struct WORDTREE;

typedef int (*wordtree_comparison)(const char* word1,
				   const char* word2);

/* comparison function used by wordtree. word1 and  */
/* word2 are always valid strings. The return value */
/* is interpreted by the wordtree as:               */
/*                                                  */
/* 0 : word1 is equal to word2                      */
/* <0: word1 is less than word2                     */
/* >0: word1 is greater than word2                  */


typedef int (*wordtree_wordvisitor)(const char* word,
				    void* userdata,
                                    void* commondata);

/* Function called for every word when traversing   */
/* the tree.                                        */
/* word is always a valid string.                   */
/* userdata is the userdata associated with the     */
/* word.                                            */
/* commondata is the pointer passed to the traverse */
/* function.                                        */
/* If the return value is 0, traversal stops.       */



typedef void (*wordtree_userdatadel)(void* userdata);

/* Function called for every userdata defined in the*/
/* tree when destroying it. The function must do    */
/* whatever is needed to take care of terminating   */
/* the userdata or resource leaks will occur.       */


WORDTREE* wordtree_create(wordtree_comparison comp);

/* Create a new wordtree.                           */
/*                                                  */
/* Return values:                                   */
/*   A valid wordtree pointer or                    */
/*   NULL on failure.                               */
/*                                                  */
/* Preconditions:                                   */
/* comp is not NULL                                 */

void** wordtree_insert(WORDTREE* tree, 
		       const char* word);

/* Insert word in tree.                             */
/*                                                  */
/* Return values:                                   */
/*   NULL on failure to insert word.                */
/*   Pointer to NULL userdata if word was inserted. */
/*   Pointer to userdata if word was already in     */
/*     the tree.                                    */
/*                                                  */
/* Preconditions:                                   */
/*   tree != NULL                                   */
/*   word != NULL                                   */


void** wordtree_find(WORDTREE* tree, const char* word);

/* Search for word in the tree.                     */
/*                                                  */
/* Return values:                                   */
/*   NULL if the word was not in the tree.          */
/*   Otherwise pointer to user data pointer         */
/*   associated with the word                       */
/*                                                  */
/* Preconditions:                                   */
/*   tree != NULL                                   */
/*   word != NULL                                   */


int wordtree_traverse(WORDTREE* tree,
		      wordtree_wordvisitor visitor,
                      void* commondata);

/* Visit word by word in the tree, in their sorted  */
/* order, calling visitor with the word and the     */
/* userdata associated with it plus the common data */
/* for all words, until all words have been visited,*/
/* or visitor returns 0                             */
/*                                                  */
/* Return values:                                   */
/*   1: Tree traversed successfully.                */
/*   0: Traversal aborted by visitor                */
/*                                                  */
/* Preconditions:                                   */
/*   tree != NULL                                   */
/*   visitor != NULL                                */


void wordtree_destroy(WORDTREE* tree,
		      wordtree_userdatadel deletor);

/* Destroy the tree, freeing all allocated memory,  */
/* and call deletor, if not NULL, for the userdata  */
/* associated with every word in the tree.          */
/*                                                  */
/* Preconditions:                                   */
/*   tree != NULL                                   */

